/*
 * Copyright 2014-2015 the original author or authors.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.schildbach.wallet.payments;

import android.os.Handler;
import android.os.Looper;

import org.bitcoinj.protocols.payments.PaymentProtocol;
import org.dash.wallet.common.util.Constants;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;

import javax.annotation.Nullable;

import de.schildbach.wallet.data.PaymentIntent;
import de.schildbach.wallet.ui.util.InputParser;
import de.schildbach.wallet_test.R;
import okhttp3.CacheControl;
import okhttp3.Call;
import okhttp3.Request;
import okhttp3.Response;

/**
 * @author Andreas Schildbach
 */
public abstract class RequestPaymentRequestTask {
    private final Handler backgroundHandler;
    private final Handler callbackHandler;
    private final ResultCallback resultCallback;

    private static final Logger log = LoggerFactory.getLogger(RequestPaymentRequestTask.class);

    public interface ResultCallback {
        void onPaymentIntent(PaymentIntent paymentIntent);

        void onFail(Exception ex, int messageResId, Object... messageArgs);
    }

    public RequestPaymentRequestTask(final Handler backgroundHandler, final ResultCallback resultCallback) {
        this.backgroundHandler = backgroundHandler;
        this.callbackHandler = new Handler(Looper.myLooper());
        this.resultCallback = resultCallback;
    }

    public final static class HttpRequestTask extends RequestPaymentRequestTask {
        @Nullable
        private final String userAgent;

        public HttpRequestTask(final Handler backgroundHandler, final ResultCallback resultCallback,
                               @Nullable final String userAgent) {
            super(backgroundHandler, resultCallback);

            this.userAgent = userAgent;
        }

        @Override
        public void requestPaymentRequest(final String url) {
            super.backgroundHandler.post(new Runnable() {
                @Override
                public void run() {
                    log.info("trying to request payment request from {}", url);

                    final Request.Builder requestBuilder = new Request.Builder()
                            .url(url)
                            .cacheControl(new CacheControl.Builder().noCache().build())
                            .header("Accept", PaymentProtocol.MIMETYPE_PAYMENTREQUEST);
                    if (userAgent != null)
                        requestBuilder.header("User-Agent", userAgent);

                    final Call call = Constants.INSTANCE.getHTTP_CLIENT().newCall(requestBuilder.build());
                    try {
                        final Response response = call.execute();
                        if (response.isSuccessful()) {
                            final String contentType = response.header("Content-Type");
                            final InputStream is = response.body().byteStream();
                            new InputParser.StreamInputParser(contentType, is) {
                                @Override
                                protected void handlePaymentIntent(final PaymentIntent paymentIntent) {
                                    log.info("received {} via http", paymentIntent);

                                    onPaymentIntent(paymentIntent);
                                }

                                @Override
                                protected void error(Exception x, final int messageResId, final Object... messageArgs) {
                                    onFail(x, messageResId, messageArgs);
                                }
                            }.parse();
                            is.close();
                        } else {
                            final int responseCode = response.code();
                            final String responseMessage = response.message();

                            log.info("got http error {}: {}", responseCode, responseMessage);
                            onFail(null, R.string.error_http, responseCode, responseMessage);
                        }
                    } catch (final IOException x) {
                        log.info("problem sending", x);

                        onFail(x, R.string.error_io, x.getMessage());
                    }
                }
            });
        }
    }

    public abstract void requestPaymentRequest(String url);

    protected void onPaymentIntent(final PaymentIntent paymentIntent) {
        callbackHandler.post(new Runnable() {
            @Override
            public void run() {
                resultCallback.onPaymentIntent(paymentIntent);
            }
        });
    }

    protected void onFail(final Exception ex, final int messageResId, final Object... messageArgs) {
        callbackHandler.post(new Runnable() {
            @Override
            public void run() {
                resultCallback.onFail(ex, messageResId, messageArgs);
            }
        });
    }
}
